CREATE TABLE public.sentinel_bursts
(
    ogc_fid integer NOT NULL DEFAULT nextval('sentinel_bursts_ogc_fid_seq'::regclass),
    "granuleUR" character varying COLLATE pg_catalog."default",
    "insertDate" timestamp with time zone,
    "productionDateTime" timestamp with time zone,
    identifier character varying COLLATE pg_catalog."default",
    "inputGranule1" character varying COLLATE pg_catalog."default",
    "beginningDateTime" timestamp with time zone,
    "endingDateTime" timestamp with time zone,
    "platformShortName" character varying COLLATE pg_catalog."default",
    "urlOfFrame" character varying COLLATE pg_catalog."default",
    swath character varying COLLATE pg_catalog."default",
    "startLineInSwath" integer,
    "endLineInSwath" integer,
    ipf character varying COLLATE pg_catalog."default",
    "numberOfLines" integer,
    "numberOfSamples" integer,
    "startingRange" real,
    "sensingStart" timestamp with time zone,
    "sensingStop" timestamp with time zone,
    "burstStartUTC" timestamp with time zone,
    "burstStopUTC" timestamp with time zone,
    "trackNumber" integer,
    "frameNumber" integer,
    "orbitNumber" integer,
    "swathNumber" integer,
    "burstNumber" integer,
    "passDirection" character varying COLLATE pg_catalog."default",
    "azimuthSteeringRate" real,
    "rangePixelSize" real,
    "rangeSamplingRate" real,
    "azimuthTimeInterval" real,
    "radarWavelength" real,
    polarization character varying COLLATE pg_catalog."default",
    "terrainHeight" real,
    prf real,
    "firstValidLine" integer,
    "numValidLines" integer,
    "firstValidSample" integer,
    "numValidSamples" integer,
    "rangeWindowType" character varying COLLATE pg_catalog."default",
    "rangeWindowCoefficient" real,
    "azimuthWindowType" character varying COLLATE pg_catalog."default",
    "azimuthWindowCoefficient" real,
    "azimuthFMRate" character varying COLLATE pg_catalog."default",
    doppler character varying COLLATE pg_catalog."default",
    "absoluteBurstID" character varying COLLATE pg_catalog."default",
    "relativeBurstID" character varying COLLATE pg_catalog."default",
    "previousBurstID" character varying COLLATE pg_catalog."default",
    "nextBurstID" character varying COLLATE pg_catalog."default",
    "absoluteESAburstID" integer,
    "relativeESAburstID" integer,
    "timeSinceAnxNode" real,
    "ascendingNodeTime" timestamp with time zone,
    "hasLand" boolean,
    "crossesDateLine" boolean,
    "landFraction" real,
    "pointLongitude_1" real,
    "pointLatitude_1" real,
    "pointLongitude_2" real,
    "pointLatitude_2" real,
    "pointLongitude_3" real,
    "pointLatitude_3" real,
    "pointLongitude_4" real,
    "pointLatitude_4" real,
    "pointLongitude_5" real,
    "pointLatitude_5" real,
    geometry geometry(Polygon,4326),
    CONSTRAINT sentinel_bursts_pkey PRIMARY KEY (ogc_fid)
)

TABLESPACE pg_default;

ALTER TABLE public.sentinel_bursts
    OWNER to postgres;
CREATE INDEX sentinel_bursts_wkb_geometry_geom_idx
    ON public.sentinel_bursts USING gist
    (geometry)
    TABLESPACE pg_default;
