CREATE SEQUENCE public.sentinel_bursts_ogc_fid_seq
    INCREMENT 1
    START 1
    MINVALUE 1
    MAXVALUE 9223372036854775807
    CACHE 1;

ALTER SEQUENCE public.sentinel_bursts_ogc_fid_seq
    OWNER TO postgres;
